# -*- coding: utf-8 -*-
#
# dsbench/tpcbBuf.py - Residence time buffer
#
# Copyright (C) 2014-2015 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

import numpy
import struct
import threading

from tpcbDefs import *


class tpcbBuf():
    """Transaction residence time buffer.

    This class delivers a dynamic buffer using numpy arrays to store
    residence times of each transaction. It has to handle millions of single
    measurements.
    """
    def __init__(self, blocksize, confidences):
        """Initialisation.

        Initialises the dynamic buffer.

        Args:
            blocksize: A number that defines one block for dynamic
                numpy array allocation.
            confidences: A configuration tuple containing the confidence
                levels for which results should be calculated.
        """
        self.blocks = 1
        self.blocksize = blocksize
        self.confidences = confidences
        self.data = numpy.zeros(self.blocksize, dtype='f')
        self.sorted = False
        self.lock = threading.Lock()
        self.nr = 0

    def __bool__(self):
        """Checks if the buffer contains values.

        Return:
            A boolean True if values are present.
        """
        return self.nr > 0

    def clean(self):
        """Clean buffer.

        Deletes all data and restores the starting conditions.
        """
        del self.data
        if hasattr(self, "datasort"):
            del self.datasort
        self.blocks = 1
        self.data = numpy.zeros(self.blocksize, dtype='f')
        self.sorted = False
        self.nr = 0

    def set(self, x):
        """Adds a single value to the buffer.

        This method is thread safe.

        Args:
            x: A single transaction residence time in seconds.
        """
        self.lock.acquire()
        try:
            if self.nr == self.data.size:
                self.blocks += 1
                self.data.resize(self.blocks * self.blocksize)
            self.data[self.nr] = x
            self.nr += 1
        finally:
            self.lock.release()
            self.sorted = False

    def setblock(self, bytearry):
        """Enters many values.

        Imports a byte array to the buffer. It is used when data are
        processed by remote nodes and transferred to the master.

        Args:
            bytearry: A byte array containing transaction residence times
                in seconds.
        """
        if bytearry is None:
            return
        data = numpy.frombuffer(bytearry, dtype='f')

        self.lock.acquire()
        try:
            if self.nr + data.size >= self.data.size:
                self.blocks = (self.data.size + data.size) // self.blocksize
                self.blocks += 1
                self.data.resize(self.blocks * self.blocksize)
            x_index = self.nr + numpy.arange(data.size)
            self.data[x_index] = data
            self.nr += data.size
        finally:
            self.lock.release()
            self.sorted = False

    def tobytes(self):
        """Output as byte array.

        Converts the buffer to a bytearray suitable for network
        transports.

        Returns:
            A byte array containing the whole buffer.
        """
        # packed = [struct.pack('f', val) for val in self.data]
        # return b''.join(packed)
        return struct.pack('%sf' % self.nr, *self.data[:self.nr])
        # return self.data.tobytes() # TODO: for numpy v1.9

    def savetxt(self, filename):
        """Save data to file.

        Saves the raw residence time data unsorted to a file.

        Args:
            filename: Name of the file to be stored.
        """
        numpy.savetxt(filename, self.data[:self.nr])

    def loadtxt(self, filename):
        """Load data from file.

        Loads the raw residence time data from a file and adapt
        internal structures.

        Args:
            filename: Name of the file.
        """
        self.clean()
        self.lock.acquire()
        try:
            self.data = numpy.loadtxt(filename)
            self.nr = self.data.size
            print(self.nr)
            self.blocks = self.nr // self.blocksize
            self.blocks += 1
            self.data.resize(self.blocks * self.blocksize)
        finally:
            self.lock.release()

    def sort(self):
        """Sort.

        As a part of statistical analysis this method sorts the data
        in the buffer.

        Returns:
            A numpy.array containing the transaction residence times sorted.
        """
        if not self.sorted:
            x_index = numpy.argsort(self.data[:self.nr])
            self.datasort = self.data[x_index]
            self.sorted = True
        return self.datasort

    def redindex(self):
        """Reducing data.

        This routine reduces the amount of data for plotting purposes by
        omitting values at a polynomic cumulative frequency scale.

        Returns:
            An index list containing the numbers of sorted transaction
            residence times selected for cumulative frequency diagram
            representation.
        """
        self.sort()

        x_index = []
        if self.datasort.size <= REDUCEOFFSET:
            # Very small amounts of data: Select all
            for i in range(self.datasort.size):
                x_index.append(i)
        else:
            # Select all large residence times of low probability
            for i in range(REDUCEOFFSET):
                x_index.append(i)
            skip = 1
            position = REDUCEOFFSET - 1
            # As shorter the residence time, i.e. as higher the probability
            # as more values will be wasted (by rising the 'skip' variable)
            while position + skip < self.datasort.size:
                position += skip
                x_index.append(position)
                skip += 2
            x_index.append(self.datasort.size - 1)

        # Revert the generated index
        for i in range(len(x_index)):
            x_index[i] = self.datasort.size - 1 - x_index[i]
        return x_index[::-1]

    def minmax(self):
        """Delivering extreme values.

        This function returns the minimum and the maximum value
        of the registered transaction residence times.

        Returns:
            A pair of floats containing minimum and maximum.
        """
        if self:
            self.sort()
            return self.datasort[0], self.datasort[-1]
        else:
            return None

    def statistics(self, multi):
        """Statistics.

        Calculates the transaction rates for a list of configured
        confidence levels (for instance for 68%, 95% and 99%).

        Args:
            multi: Multiplicator for transaction rates calculated
                from total residence time and total demanded duration.

        Returns:
            A list of transaction rates in seconds for each confidence
            level in the same order as the levels.
        """
        if self:
            self.sort()
            retval = []
            for c in self.confidences:
                idx = int(self.nr * c / 100.0)
                tpcb = multi / self.datasort[idx]
                retval.append(tpcb)
            return retval
        else:
            return None


if __name__ == "__main__":
    """Module testing.

    This main part is used for testing the module.
    """
    import timeit
    import random

    size = 100000
    buf = tpcbBuf(size, 1)

    for u in range(21):
        t1 = timeit.default_timer()
        n = 99999
        for i in range(n):
            buf.set(random.random() * 20.0 + 1)
        t2 = timeit.default_timer()
        print(n, "x set(value):", t2 - t1)

    t1 = timeit.default_timer()
    d = buf.sort()
    t2 = timeit.default_timer()
    print("unsort:", buf.data, buf.data.size)
    print("sort:", d, d.size, t2 - t1)

    t1 = timeit.default_timer()
    b = buf.tobytes()
    t2 = timeit.default_timer()
    print("tobytes:", b[:20], len(b), t2 - t1)

    t1 = timeit.default_timer()
    buf.setblock(b)
    t2 = timeit.default_timer()
    print("setblock:", buf.data, buf.data.size, t2 - t1)

    t1 = timeit.default_timer()
    d = buf.sort()
    t2 = timeit.default_timer()
    print("sort:", d, d.size, t2 - t1)
