# -*- coding: utf-8 -*-
#
# dsbench/tpcbDbBase.py - Database core
#
# Copyright (C) 2014-2015 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

import queue
import timeit
import socketserver

from tpcbCore import *


# i18n preparations.
def _(s):
    return s


# Constants
OUTRESULT = "cycle={0:d}, scale={1:d}, dbsize={2:f}, done={3:d}, " \
            "errors={4:d}, residence={5:.2f}, elapsed={6:.2f}, " \
            "tpsR={7:.2f}, tpsE={8:.2f}, tpsD={9:.2f}"


class tpcbMaster(socketserver.ThreadingTCPServer):
    """TPC master server.

    The master TCP server is responsible for listening for node connects
    as reponse to a UDP broadcast. The server delivers transaction jobs
    to the nodes connected and wait for their responses that contain
    mesurement results.
    """
    def __init__(self, remote, addr, handler):
        """Initialisation.

        Args:
            remote: tpcbRemote() instance, the master control thread
                responsible for controlling all remote nodes
            addr: Address and port the TPC server should listen to.
            handler: A tpcbMasterHandler() class that act as threaded
                handler if a remote nodes connects to the master.
        """
        socketserver.ThreadingTCPServer.__init__(self, addr, handler)
        self.remote = remote

    def server_bind(self):
        """Bind the server.

        Overrides the binding process to the server socket. Important
        here: Option socket.SO_REUSEADDR, that tells the system to release
        socket resources after closing.
        """
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.server_address)

    def serve_forever(self):
        """Main loop of the server.

        Overrides the main loop of the server socket. Besides the normal
        loop it sets the event 'started' to signal the main thread that the
        TCP server is up now. The time needed between setting the event
        'started' and the real start of the event handler loop in
        super().serve_forever() is considered by a delay. See also notes
        in tpcbRemote.waitstart()!
        """
        self.remote.started.set()
        # socketserver.ThreadingTCPServer.serve_forever()
        super().serve_forever()


class tpcbMasterHandler(socketserver.BaseRequestHandler):
    """Handler for TCP connections.

    If a node connects to the master, the handle() method of this
    request handler class is invoked in a separate thread. Each handler
    thread manages exactly one connected node.
    """
    def close(self):
        """Close connection to the node."""
        self.logger.debug(_("Closing connection to node '%s'."),
                          self.client_address[0])
        self.request.close()

    def warn(self, msg, *args):
        """Handle warning.

        This method gives a warn message to the logger and closes the
        connection to the node. This is used for instance for nodes that
        are connected but not required for measurements.

        Args:
            msg: Message string to be logged.
            *args: Positional arguments to be inserted in the message.
        """
        self.logger.warn(msg, *args)
        self.close()

    def error(self, msg, *args):
        """Handle error.

        This method gives an error message and closes the connection
        to the client. This kind of errors are critical and prevent
        continuation of the measurement process. This is signalised by
        a flag and a finish event.

        Args:
            msg: Message string to be logged.
            *args: Positional arguments to be inserted in the message.
        """
        self.logger.error(msg, *args)
        self.close()
        self.remote.error = True
        self.remote.finished.set()

    def handle(self):
        """Node handler.

        Override method that is started as thread if a remote node
        connects to the master TCP server. This method expects an
        identification of the remote node and delivers a transaction job.
        Then it waits until a result is delivered by the remote node.
        """
        # Initialisation
        self.remote = self.server.remote
        self.remote.handled.set()
        env = self.remote.db.env
        single = self.remote.db.single
        self.logger = env.logger
        self.remote.lock.acquire()
        try:
            self.remote.tnr += 1
            threading.current_thread().name = "Manager-{0:02d}".format(
                self.remote.tnr)
        finally:
            self.remote.lock.release()

        self.logger.debug(_("Node '%s' connected."), self.client_address[0])

        # Ignore all nodes if requested number is still gathered
        if self.remote.availableCount == 0:
            self.warn(_("Already found enough nodes. Ignore node '%s'."),
                      self.client_address[0])
            return

        # Expect message from node for identification
        # self.request is the TCP socket connected to the node
        self.request.settimeout(TIMEOUT)
        buffer = self.request.recv(BUFSIZE)
        if not buffer:
            self.warn(_("Connection closed unexpectedly. Ignore node '%s'."),
                      self.client_address[0])
            return
        msg = tpcbMsg(buffer=buffer)
        self.logger.debug(MSGFROMNODE, self.client_address[0], str(msg))

        # Ignore all except correct identification
        if not msg.isnode:
            self.warn(
                _("Message is not a valid identification. Ignore node '%s'."),
                self.client_address[0])
            return

        # Register remote node
        if not env.registernode(msg.parameters, self.client_address[0]):
            self.warn(_("Node not registered. Reject node '%s'."),
                      self.client_address[0])
            return

        # If enough nodes are collected an event is invoked that
        # signalises all active handlers to continue (see below).
        self.remote.lock.acquire()
        try:
            self.remote.availableCount -= 1
            if self.remote.availableCount == 0:
                self.remote.available.set()
        finally:
            self.remote.lock.release()

        # Waiting for start signal
        # Stop processing here until the required number of nodes
        # are available. This is signalised by an event (see above).
        # A timeout prevents hanging if not enough nodes are available.
        # If an other nodes detects an error the connection is closed
        # immediately.
        if not self.remote.available.wait(TIMEOUT):
            self.error(_("Not enough nodes available to continue."))
            return
        else:
            if self.remote.error:
                self.warn(_("Close connection to node '%s' due to error."),
                          self.client_address[0])
                return

        # Send a job
        msg = tpcbMsg(msgtype="JOB", parameters=str(env.job))
        self.logger.debug(_("Send message to node '%s': '%s'."),
                          self.client_address[0], str(msg))
        self.request.sendall(bytes(msg))

        # Wait for response (timeout determined by duration)
        # within a loop until an end mark is received because due to
        # many threads the result string can be fragmented.
        res = tpcbMsg()
        self.request.settimeout(env.job.timeout)
        while True:
            buffer = self.request.recv(BUFSIZE)
            if not buffer:
                self.error(ERRNODECLOSED, self.client_address[0])
                return

            msg = tpcbMsg(buffer=buffer)
            if msg.isendmark:
                self.logger.debug(_("Endmark received from node '%s'."),
                                  self.client_address[0])
                break
            else:
                res.append(buffer)
                self.logger.debug(MSGFROMNODE,
                                  self.client_address[0], str(msg))

            self.request.settimeout(TIMEOUT)

        # Ignore all messages except correct results
        if not res.isresult:
            self.error(_("Message from node '%s' is not a valid result."),
                       self.client_address[0])
            return

        # Parse results
        results = eval(res.parameters)

        # Send received results to logging module
        for item in results:
            out = OUTRESULT.format(
                env.job.cycle, env.job.scale, env.job.dbsize,
                item.done, item.errors, item.residence, item.elapsed,
                item.tpsR, item.tpsE, item.tpsD)
            self.logger.info(out)

        # All next received packages are transaction residence times
        # as byte converted numpy.ndarray containing floats or an end
        # mark
        self.request.settimeout(TIMEOUT)
        while True:
            buffer = self.request.recv(BUFSIZE)
            if not buffer:
                self.error(ERRNODECLOSED, self.client_address[0])
                return
            msg = tpcbMsg(buffer=buffer)
            if msg.isendmark:
                break
            # -------------------------------------------------------
            # TODO: Sometimes data is missing.
            if msg.parameters is None:
                message = _("Missing suitable data from node '{0:s}'.").format(
                    self.client_address[0])
                self.logger.error(message)
                self.logger.error(_("Message:\n") + str(msg) + "\n")
            # -------------------------------------------------------
            single.buf.setblock(msg.parameters)

        # Save results (see above)
        self.remote.lock.acquire()
        try:
            self.remote.finishCount -= 1
            for item in results:
                self.remote.results.append(item)
            if self.remote.finishCount == 0:
                self.remote.finished.set()
        finally:
            self.remote.lock.release()

        # Close connection
        self.close()


class tpcbRemote(threading.Thread):
    """Master controller.

    Master thread that controls all remote node connections.
    """
    def __init__(self, db):
        """Initialisation.

        Args:
            db: An instance of tpcbDbBase() class or related containing
                the database.
        """
        threading.Thread.__init__(self)
        self.started = threading.Event()
        self.handled = threading.Event()
        self.available = threading.Event()
        self.finished = threading.Event()
        self.lock = threading.Lock()
        self.name = "Master"
        self.db = db
        self.reset()

    def shutdown(self):
        """Shutdown.

        Aborts the thread by shutting down the server and wait until
        master thread finished.
        """
        self.server.shutdown()
        self.join()

    def waitstart(self):
        """Waiting for start.

        Wait for the event 'started' which is set if the TCP server is
        started successfully. This is required to avoid that the main thread
        sends UDP broadcasts before the TCP server responsible to fetch
        responses of the nodes is running. After receiving the event
        'started' a delay of a second ensures that the handler loop of
        the TCP server is also invoked. See also notes in
        tpcbMaster.serve_forever()!
        """
        if not self.started.wait(TIMEOUT):
            msg = _("Abort waiting for TCP server startup due to timeout.")
            raise Exception(msg)
        self.db.env.logger.debug(_("TCP server ready to listen for nodes."))
        time.sleep(1)

    def reset(self):
        """Reset instance.

        Restores the start conditions for a new transaction cycle
        attempt.
        """
        self.tnr = 0
        self.results = []
        self.error = False
        self.handled.clear()
        self.available.clear()
        self.finished.clear()
        self.lock.acquire()
        try:
            self.availableCount = self.db.env.job.nodes
            self.finishCount = self.db.env.job.nodes
        finally:
            self.lock.release()

    def broadcast(self):
        """UDP broadcast.

        Sends a UDP broadcast to wake up nodes in the LAN. This is the
        first step to send transaction jobs to remote nodes.
        """
        self.db.env.logger.debug(_("Send UDP broadcast to find nodes."))
        s = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, True)
        msg = tpcbMsg(msgtype="SCAN")
        s.sendto(bytes(msg), ("<broadcast>", self.db.env.port))
        s.close()

    def waitresults(self):
        """Getting results.

        Wait at first for the event 'handled' which is set if the first
        node is found and then for the event 'finished' which is set
        if all nodes delivered valid results. A timeout is used to avoid
        hanging if a handler crashed unexpectedly. In case of node errors
        this routine tries to start again.
        """
        ntrials = max(self.db.env.ntrials, 1)

        while ntrials > 0:
            # Prepare a new attempt
            self.reset()

            # Send broadcast to wake up nodes
            self.broadcast()

            # Wait for first successful node response
            if not self.handled.wait(TIMEOUT):
                msg = _("Abort waiting for node connects due to timeout.")
                raise Exception(msg)
            self.db.env.logger.debug(
                _("First node signalised ready to start transactions."))

            # Wait for results or for error signals from nodes
            if not self.finished.wait(self.db.env.job.timeout):
                msg = _("Abort waiting for node result delivery due to "
                        "timeout.")
                raise Exception(msg)
            else:
                if self.error:
                    # Node error: Try it again
                    self.db.env.logger.warn(
                        _("Error in node handling detected, try it again."))
                    ntrials -= 1
                    continue
                else:
                    # Successfully received node results
                    return

        # Configured number of trials failed: Throw an exception
        msg = _("Aborted due to repeated error in node handling.")
        raise Exception(msg)

    def run(self):
        """Master thread code.

        Installs the TCP server that waits for node connections and starts
        the listening loop in a separate thread.
        """
        addr = ("", self.db.env.port)
        self.db.env.logger.debug(
            _("Install TCP server listening on '%s'."), addr)
        self.server = tpcbMaster(self, addr, tpcbMasterHandler)
        self.server.serve_forever()
        del self.server

    def result(self):
        """Pickup results.

        Returns:
            The measurement results as a list of tpcbResults() tuples.
        """
        return self.results


class tpcbThread(threading.Thread):
    """Threaded transaction processing.

    This class controls a thread to execute transactions on the database.
    """
    def __init__(self, db, q, tnr):
        """Initialisation:

        Args:
            db: An instance of tpcbDbBase() class or related containing
                the database.
            tnr: Number of the thread for thread naming.
        """
        threading.Thread.__init__(self)
        self.evkill = threading.Event()
        self.name = "Worker-{0:02d}".format(tnr)
        self.results = None
        self.db = db
        self.q = q

    def kill(self):
        """Kill thread.

        Aborts the thread by setting an event which is checked in the
        transaction loop. This is useless if a thread hangs within a
        transaction.
        """
        self.evkill.set()

    def run(self):
        """Thread code.

        This method invokes the transaction loop of the database class
        instance. In case of an exception the results remains None to
        signalise error to the main thread.
        """
        try:
            if self.db.env.job.reconnect:
                self.results = self.db.run_single(self.evkill)
            else:
                self.results = self.db.run_multi(self.evkill)
        finally:
            self.q.put(self.results)


class tpcbDb():
    """Generic database controller.

    This is a generic base class for database operations needed for TPC-B
    transaction measurements.
    """
    def __init__(self, env, multiple):
        """Initialisation.

        Args:
            env: The test environment, a tpcbEnv() instance.
            multiple: The measurements table, a tpcbMultiple() instance.
        """
        self.env = env
        self.multiple = multiple
        self.name = _("generic")
        self.basepath = ""

    def _connect(self):
        """Connect to the database.

        This is an empty method for override.

        Returns:
            A connection and a cursor to the database.
        """
        con = None
        cur = None
        return con, cur

    def _disconnect(self, con, cur):
        """Disconnect from database.

        This is an empty method for override.

        Args:
            con: Connection to the database
            cur: Cursor from the database connection
        """
        pass

    def _create(self, scale):
        """Create a database.

        This is an empty method for override.

        Args:
            scale: Scaling factor of the database.

        Returns:
            A tuple containing durations in seconds for filling the
            database, creating the primary and foreign keys.
        """
        tInit = 0.0
        tPk = 0.0
        tFk = 0.0
        return tInit, tPk, tFk

    def _drop(self):
        """Drop the database.

        This is an empty method for override.

        Returns:
            Time needed to drop the database in seconds.
        """
        tDrop = 0.0
        return tDrop

    def _clean(self):
        """Clean the database between measurement cycles.

        This is an empty method for override.

        Returns:
            Time needed to clean the database in seconds.
        """
        tClean = 0.0
        return tClean

    def _size(self):
        """Determine the physical size of the database in GB.

        This is an empty method for override.

        Returns:
            The size of the database in GB as a float.
        """
        dbsize = 0.0
        return dbsize

    def _execute(self, sql, con, cur):
        """Executes SQL commands.

        This is an empty method for override.

        Args:
            sql: SQL statements.
            con: Connection to the database.
            cur: Cursor on the connection.
        """
        pass

    def _loop_single(self, evkill):
        """Transaction loop with single connection.

        This is an empty method for override.
        Each transaction opens and closes the database connection.

        Args:
            evkill: A thread event to signalise kill.

        Returns:
            A tuple containing the number of successful transactions done,
            the number of rejected transactions due to collisions, the total
            transaction residence time and effective duration elapsed during
            transactions.
        """
        done = 0
        errors = 0
        residence = 0.0
        elapsed = 0.0
        return done, errors, residence, elpased

    def _loop_multi(self, evkill, cons, curs):
        """Transaction loop with multiple connections.

        This is an empty method for override.
        Connections are opened and closed only once for the whole
        measurement cycle.

        Args:
            evkill: A thread event to signalise kill.
            cons: A list of connections to the database.
            curs: A list of cursors on the connections.

        Returns:
            A tuple containing the number of successful transactions done,
            the number of rejected transactions due to collisions, the total
            transaction residence time and effective duration elapsed during
            transactions.
        """
        done = 0
        errors = 0
        residence = 0.0
        elapsed = 0.0
        return done, errors, residence, elpased

    def _parameters(self):
        """Collect database parameters.

        This is an empty method for override.
        """
        pass

    def timeexec(self, prnt, sql, con, cur):
        """Execute SQL and measure time.

        Executes an SQL code line and measures its duration.
        Displays begin and end of execution in the logging module.

        Args:
            prnt: Line to print in the logger.
            sql: SQL statements.
            con: Connection to the database.
            cur: Cursor on the connection.

        Returns:
            Residence time of SQL execution.
        """
        self.env.logger.debug(prnt + " ... " + _("started") + ".")
        start = timeit.default_timer()
        self._execute(sql, con, cur)
        end = timeit.default_timer()
        self.env.logger.debug(prnt + " ... " + _("finished") + ".")
        return end - start

    def fileexec(self, prnt, filename, con, cur):
        """Execute SQL file and measure time.

        Executes an SQL script and displays begin and end in the
        logging module. The time required for execution is acquired.

        Args:
            prnt: Line to print in the logger.
            filename: Name of file containing SQL statements.
            con: Connection to the database.
            cur: Cursor on the connection.

        Returns:
            Residence time of SQL execution.
        """
        path = os.path.join(self.basepath, filename)
        self.env.logger.debug(_("Path to SQL file: '%s'"), path)
        sql = open(path, "r").read()
        start = timeit.default_timer()
        self._execute(sql, con, cur)
        end = timeit.default_timer()
        self.env.logger.debug(prnt)
        return end - start

    def create(self, single, database):
        """Create Database.

        Creates a database of given scaling factor and measures and
        stores concurrent times for insert and generation of primary and
        foreign keys. Determines the size of the created database.

        Args:
            single: A tpcbSingle() instance to be measured.
            database: A boolean that is True if the database should be
                created. Create is invoked with False by the remote nodes
                to install the random number generator without creating
                the database.
        """
        self.single = single
        self.env.job.scale = single.scale

        if database:
            # Logging
            self.env.blankinfo(
                _("\n=== Begin {0:s} scale run {1:d} ===").format(
                    NAME, single.scale))
            # Create database
            tinit, tpk, tfk = self._create(self.single.scale)
            self.single.setvalues(tinit=tinit)
            self.single.setvalues(tpk=tpk)
            self.single.setvalues(tfk=tfk)
            # Determine size of database
            self.env.job.dbsize = self._size()
            self.single.setvalues(dbsize=self.env.job.dbsize)

        # Initialise random generator
        self.rand = tpcbRandom(True, single.scale)

    def drop(self, database):
        """Drop database.

        Removes the database if configured and measures the time.

        Args:
            database: A boolean that is True if the database should be
                dropped. Drop is invoked with False by the remote nodes
                to uninstall the random number generator without dropping
                the database.
        """
        if database:
            # Logging
            self.env.blankinfo(
                _("=== Cleanup {0:s} scale run {1:d} ===").format(
                    NAME, self.single.scale))
            # Remove database
            if self.env.remove:
                tdrop = self._drop()
                self.single.setvalues(tdrop=tdrop)

        # Mark the data record as completed
        self.single.done()
        del self.rand

    def results(self, done, errors, residence, elapsed):
        """Interpret results.

        Calculates the transaction rates and prints the results
        to the logger.

        Args:
            done: Number of transactions done successfully.
            errors: Number of failed transactions.
            residence: Total transaction residence time.
            elapsed: Effective duration of transaction test cycle.

        Returns:
            The measurement results as a tpcbResults() tuple.
        """
        if residence > 0.0:
            tpsR = done / residence
        else:
            tpsR = 0.0

        if elapsed > 0.0:
            tpsE = done / elapsed
        else:
            tpsE = 0.0

        tpsD = done / self.env.job.duration

        out = OUTRESULT.format(
            self.env.job.cycle, self.env.job.scale, self.env.job.dbsize,
            done, errors, residence, elapsed, tpsR, tpsE, tpsD)
        self.env.logger.info(out)

        return tpcbResults(self.env.own.host,
                           done, errors, residence, elapsed,
                           tpsR, tpsE, tpsD)

    def run_single(self, evkill):
        """Transactions on single session.

        Runs transaction measurements in only one database session
        which opens and closes the connection to the database for each
        transaction.

        Args:
            evkill: A thread event to signalise kill.

        Returns:
            The measurement results as a tpcbResults() tuple.
        """
        done, errors, residence, elapsed = self._loop_single(evkill)
        return self.results(done, errors, residence, elapsed)

    def run_multi(self, evkill):
        """Transaction an multipe sessions.

        Distributes the transactions to several configured database
        sessions. Runs one cycle of TPC-B transaction rate measurements.
        Note, that this method must not change any part of self because
        several threaded instances of this method use the same data!

        Args:
            evkill: A thread event to signalise kill.

        Returns:
            The measurement results as a tpcbResults() tuple.
        """
        # Connect several database sessions
        cons = []
        curs = []
        for c in range(self.env.job.clients):
            con, cur = self._connect()
            cons.append(con)
            curs.append(cur)

        # Database specific transaction loop
        done, errors, residence, elapsed = self._loop_multi(evkill, cons, curs)
        results = self.results(done, errors, residence, elapsed)

        # Close database sessions
        for c in range(self.env.job.clients):
            self._disconnect(cons[c], curs[c])

        # Cleanup
        while curs:
            del curs[0]
        while cons:
            del cons[0]

        return results

    def kill_threads(self):
        """Killing threads.

        This routine helps the run_threads() routine to terminate
        still running threads. It waits until all threads are finished.
        If the threads are dead locked for some reasons (as observed in
        our Firebird tests) the routine kill() is unable to stop the
        thread and join() would wait endlessly. Therefore, a timeout with
        ignoring the exception is included.
        """
        for w in self.workers:
            try:
                w.kill()
                w.join(TIMEOUT)
            except:
                pass

    def run_threads(self):
        """Start several threads for transactions.

        Opens several threads with run() method, waits them and collects
        transaction rate results by means of a queue. The measured rates
        of all threads are added.

        Returns:
            The list of measurement results as tpcbResults() tuples one
            for each thread.
        """
        # Start threads
        tnr = 1
        self.workers = []
        q = queue.Queue()
        for t in range(self.env.job.threads):
            w = tpcbThread(self, q, tnr)
            self.workers.append(w)
            w.start()
            tnr += 1

        # Wait for results in a queue
        results = []
        try:
            crash = False
            # First queue entry is expected depending on duration
            timeout = self.env.job.timeout
            for w in self.workers:
                result = q.get(True, timeout)
                if result is None:
                    crash = True
                    break
                results.append(result)

            if crash:
                self.kill_threads()
                msg = _("At least one thread crashed unexpectedly. "
                        "Terminate all others.")
                raise Exception(msg)

        except queue.Empty:
            # If the queue.get() function timed out.
            self.kill_threads()
            msg = _("Only {0:d} results received but {1:d} expected.").format(
                len(results), len(self.workers))
            raise Exception(msg)

        except KeyboardInterrupt as e:
            # In case of keyboard interrupts,
            # demand thread abort and wait until finished.
            # It does not work for fdb and Firebird.
            self.kill_threads()
            raise e

        finally:
            while self.workers:
                del self.workers[0]

        return results

    def run_local(self):
        """Start transactions locally.

        Executes transactions on local node.

        Returns:
            The list of measurement results as tpcbResults() tuples one
            for each thread.
        """
        if self.env.job.threads < 1:
            # Single thread
            results = []
            if self.env.job.reconnect:
                results.append(self.run_single(None))
            else:
                results.append(self.run_multi(None))
            return results
        else:
            # Multiple threads
            return self.run_threads()

    def run_remote(self):
        """Start transactions remote.

        Connects several remote nodes and tells them to make transactions.

        Returns:
            The list of measurement results as tpcbResults() tuples one
            for each thread on each node.
        """
        w = tpcbRemote(self)
        # Install TCP server and start listening
        w.start()
        try:
            # Waiting until TCP server is up
            w.waitstart()
            # Send UDP broadcast to activate remote nodes
            # and wait until nodes delivered results
            w.waitresults()
            # Shutting down the TCP server
            w.shutdown()
            # Pick up results
            results = w.result()
        except Exception as e:
            # In case of exceptions like keyboard interrupts,
            # shutdown the server immediately and waste results
            w.shutdown()
            raise e
        finally:
            del w

        return results

    def do(self):
        """Overall measurement loop.

        Runs several transaction rate measurement cycles for the current
        database having a definite scaling factor.
        """
        tclean = 0.0
        # for i in range(self.env.no):
        while self.env.job.cycle < self.env.no:

            # Logging
            # self.env.job.cycle = i + 1
            self.env.job.cycle += 1
            self.env.blankinfo(
                _("--- Begin {0:s} measuring cycle {1:d}/{2:d} ---").format(
                    NAME, self.env.job.cycle, self.env.no))

            # Clean the database
            tclean += self._clean()

            # Start system load acquisition
            load = tpcbLoad(self.env)
            load.start()

            # Measurement cycle
            if self.env.job.nodes > 0:
                # Multiple remote nodes
                results = self.run_remote()
            else:
                # Local node
                results = self.run_local()

            # Stop IO load acquisition
            load.stop()
            self.env.logger.info(_("System load: %s"), str(load))

            # Store results
            self.single.setvalues(results=results)

        self.env.job.cycle = 0
        tclean /= self.env.no
        self.single.setvalues(tclean=tclean)

    def parameters(self):
        """Collect database parameters.

        Invokes a database specific method to collect parameters and
        configures the tpcbMultiple() measurement table for graphical
        results presentation.
        """
        self._parameters()
        self.multiple.setmarker(self.env.own.ram, "RAM")
        self.multiple.setmarker(self.env.own.ram + self.env.own.swap,
                                "RAM + SWAP")
        # List all concurrent time measurements
        self.multiple.settimedata("tinit", _("populate database"))
        self.multiple.settimedata("tpk", _("primary key"))
        self.multiple.settimedata("tfk", _("foreign key"))
        self.multiple.settimedata("tclean", _("cleanup"))
        if self.env.remove:
            self.multiple.settimedata("tdrop", _("drop database"))
        # Output a test environment representation
        self.env.logger.debug(_("Environment: %s"), repr(self.env))

    def loglast(self):
        """Logs the last measurement results."""
        self.env.logger.info(_("Results:\n%s"), repr(self.single))
