# -*- coding: utf-8 -*-
#
# dsbench/tpcbDbFb.py - Some basics for Firebird
#
# Copyright (C) 2014-2015 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

from tpcbCore import *
from tpcbDbBase import *


# i18n preparations.
def _(s):
    return s


class tpcbDbFb(tpcbDb):
    """Database controller for Firebird.

    This controller overrides tpcbDb() class method for accessing a
    Firebird database based on the 'fdb' driver.
    """
    def __init__(self, env, multiple):
        """Initialisation.

        Args:
            env: The test environment, a tpcbEnv() instance.
            multiple: The measurements table, a tpcbMultiple() instance.
        """
        tpcbDb.__init__(self, env, multiple)
        self.server = env.job.fbServer
        self.dsn = "{0:s}:tpc.fdb".format(env.job.fbServer)
        self.dbname = "tpc.fdb"
        self.user_name = "SYSDBA"
        self.password = env.job.fbPassword
        self.name = "Firebird"
        self.basepath = "firebird"
        self.config = env.job.fbConfig

    def readconfig(self, pagesize):
        """Reads Firebird configuration.

        This method extracts some important information about the
        Firebird RDBMS from the configuration file.

        Args:
            pagesize: The database page size is required to convert
                configurated sizes to GB.

        Returns:
            A tuple of parameters.
        """
        databaseAccess = ""
        defaultDbCachePages = 0
        fileSystemCacheThreshold = 0

        file = None
        try:
            file = open(self.config, mode="r")
            for line in file:
                # Ignore comments
                if not line.startswith("#"):
                    # Parse definite configuration lines
                    v = lambda s: s.split("=")[1].strip()
                    reduced = line.strip()
                    if reduced.startswith("DefaultDbCachePages"):
                        defaultDbCachePages = int(v(reduced))
                    elif reduced.startswith("FileSystemCacheThreshold"):
                        fileSystemCacheThreshold = int(v(reduced))
                    elif reduced.startswith("DatabaseAccess"):
                        value = v(reduced)
                        if value.startswith("Restrict"):
                            databaseAccess = value.replace(
                                "Restrict", "").strip()

            # Convert values
            defaultDbCachePages = defaultDbCachePages * pagesize / GB
            fileSystemCacheThreshold = fileSystemCacheThreshold * pagesize / GB

        except Exception as e:
            self.env.logger.error(
                _("Could not read Firbird configuration file '%s'. "
                  "Error message: '%s'"), self.config, str(e))
        finally:
            if file:
                file.close()

        return databaseAccess, defaultDbCachePages, fileSystemCacheThreshold
