# -*- coding: utf-8 -*-
#
# dsbench/tpcbGUI.py - Configuration window
#
# Copyright (C) 2014-2015 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

import re
import tkinter

from tpcbDefs import *


# i18n preparations.
def _(s):
    return s


class tpcbWin():
    """GUI for test conditions.

    This class contains a graphical user interface that enables to change
    test conditions. From the test environment the following parameters
    will be asked:
    - database: Radiobutton
    - transaction type: Radiobutton
    - file name for results: Entry
    - number of nodes: Entry (integers)
    - number of threads: Entry (integers)
    - number of clients: Entry (integers)
    - duration: Entry (floats)
    - number of measurement cycles: Entry (integers)
    - list of scaling factors: Entry (several integers)
    - prepare transaction: Checkbutton
    - remove database: Checkbutton
    - reconnect for each transaction: Checkbutton
    - retry failed transaction: Checkbutton
    """
    def keyPressed(self, event):
        """Reset error signalisation.

        Reset the highlight color and unbind the event handler.

        Args:
            event: Event class containing information about calling
                instance.
        """
        event.widget["highlightcolor"] = "#000000"
        event.widget.unbind("<KeyPress>")

    def back(self, entry):
        """Error signalisation.

        Sets the focus back to the errorneous entry field and colors
        the borders red to signalise the input error. An event handler
        is installed to reset the highlight color when the first key
        is pressed.

        Args:
            entry: The entry that contains errorneous inputs.
        """
        entry.delete(0, tkinter.END)
        entry["highlightcolor"] = "#FF0000"
        entry.bind("<KeyPress>", self.keyPressed)
        entry.focus_set()

    def ok(self):
        """OK pressed.

        This button pressed event handler reads out and parses all user
        inputs and close the window.
        """
        self.db = self.svDb.get()
        self.transtype = self.svTransType.get()
        self.output = self.enOutput.get()

        try:
            nodes = int(self.enNodes.get())
        except:
            self.back(self.enNodes)
            return
        self.nodes = nodes

        try:
            threads = int(self.enThreads.get())
        except:
            self.back(self.enThreads)
            return
        self.threads = threads

        try:
            clients = int(self.enClients.get())
        except:
            self.back(self.enClients)
            return
        self.clients = clients

        try:
            duration = float(self.enDuration.get())
        except:
            self.back(self.enDuration)
            return
        self.duration = duration

        try:
            no = int(self.enNo.get())
        except:
            self.back(self.enNo)
            return
        self.no = no

        scales = []
        values = self.enScales.get().strip()
        items = re.split(r" |,|;|:", values)
        try:
            for item in items:
                item = item.strip()
                if len(item) > 0:
                    scales.append(int(item))
        except:
            self.back(self.enScales)
            return
        self.scales = scales

        if self.ivPrepare.get():
            self.prepare = True
        else:
            self.prepare = False

        if self.ivRemove.get():
            self.remove = True
        else:
            self.remove = False

        if self.ivReconnect.get():
            self.reconnect = True
        else:
            self.reconnect = False

        if self.ivRetry.get():
            self.retry = True
        else:
            self.retry = False

        self.retval = 2
        self.main.destroy()

    def ignore(self):
        """IGNORE pressed.

        This button pressed event handler ignores all user inputs and
        closes the window.
        """
        self.retval = 1
        self.main.destroy()

    def cancel(self):
        """CANCEL pressed.

        This button pressed event handler ignores all user inputs and
        closes the window too.
        """
        self.retval = 0
        self.main.destroy()

    def __init__(self, db, transtype,
                 output, nodes, threads, clients, duration, no,
                 scales, prepare, remove, reconnect, retry):
        """Initialisation.

        Hand over all default values from caller.

        Args:
            db: Database shortcut.
            transtype: Transaction type.
            output: File name core for measurement results.
            nodes: Number of remote nodes to process transactions.
            threads: Number of threads per node.
            clients: Number of connections per thread.
            duration: Duration of one measurement cycle in seconds.
            no: Number of measurement cycles.
            scales: List of database scaling factors.
            prepare: Use prepare for transaction processing command.
            remove: Remove database after measurements.
            reconnect: Reconnect the database for each transaction.
            retry: Retry failed transactions until success.
        """
        self.retval = 0
        self.db = db
        self.transtype = transtype
        self.output = output
        self.nodes = nodes
        self.threads = threads
        self.clients = clients
        self.duration = duration
        self.no = no
        self.scales = scales
        self.prepare = prepare
        self.remove = remove
        self.reconnect = reconnect
        self.retry = retry

    def show(self):
        """Open window.

        Creates and opens the window and starts the event loop.

        Returns:
            Delivers an integer:
                0 - Cancel pressed
                1 - Ignore pressed
                2 - OK pressed
        """
        self.main = tkinter.Tk()
        self.main.title(_("Define Test Conditions"))

        # Entries
        # Database selection
        row = 0
        entrywidth = 30
        lbDatabase = tkinter.Label(
            self.main, text=_("Database"))
        lbDatabase.grid(row=row, column=0, sticky="w")

        self.svDb = tkinter.StringVar()
        self.svDb.set(self.db)
        cbDbPg = tkinter.Radiobutton(
            self.main, text="PostgreSQL", variable=self.svDb, value="PG")
        cbDbPg.grid(row=row, column=1)
        cbDbFb = tkinter.Radiobutton(
            self.main, text="Firebird", variable=self.svDb, value="FB")
        cbDbFb.grid(row=row, column=2)
        # cbDbFb = tkinter.Radiobutton(
        #    self.main, text="Oracle", variable=self.svDb, value="ORA")
        # cbDbFb.grid(row=row, column=3)

        # Transaction type
        row += 1
        lbTransType = tkinter.Label(self.main, text=TRANSTYPE)
        lbTransType.grid(row=row, column=0, sticky="w")

        self.svTransType = tkinter.IntVar()
        self.svTransType.set(self.transtype)

        count = 0
        check = False
        for key in tpcbTransTable.keys():
            if check and count == 0:
                row += 1
            cbTransType = tkinter.Radiobutton(
                self.main, text=tpcbTransTable[key].descr,
                variable=self.svTransType, value=key)
            cbTransType.grid(row=row, column=count + 1)
            count = (count + 1) % 3
            check = True

        # File name for output
        row += 1
        lbOutput = tkinter.Label(
            self.main, text=_("Output file name"))
        lbOutput.grid(row=row, column=0, sticky="w")

        self.enOutput = tkinter.Entry(
            self.main, width=entrywidth)
        self.enOutput.grid(row=row, column=1, columnspan=3)
        self.enOutput.insert(0, self.output)

        # Number of nodes
        row += 1
        lbNodes = tkinter.Label(self.main, text=NUMBERNODES)
        lbNodes.grid(row=row, column=0, sticky="w")

        self.enNodes = tkinter.Entry(
            self.main, width=entrywidth)
        self.enNodes.grid(row=row, column=1, columnspan=3)
        self.enNodes.insert(0, str(self.nodes))

        # Number of threads
        row += 1
        lbThreads = tkinter.Label(self.main, text=NUMBERTHREADS)
        lbThreads.grid(row=row, column=0, sticky="w")

        self.enThreads = tkinter.Entry(
            self.main, width=entrywidth)
        self.enThreads.grid(row=row, column=1, columnspan=3)
        self.enThreads.insert(0, str(self.threads))

        # Number of clients
        row += 1
        lbClients = tkinter.Label(self.main, text=NUMBERCLIENTS)
        lbClients.grid(row=row, column=0, sticky="w")

        self.enClients = tkinter.Entry(
            self.main, width=entrywidth)
        self.enClients.grid(row=row, column=1, columnspan=3)
        self.enClients.insert(0, str(self.clients))

        # Duration of transaction test cycles
        row += 1
        lbDuration = tkinter.Label(
            self.main, text=_("Duration of transactions"))
        lbDuration.grid(row=row, column=0, sticky="w")

        self.enDuration = tkinter.Entry(
            self.main, width=entrywidth)
        self.enDuration.grid(row=row, column=1, columnspan=3)
        self.enDuration.insert(0, str(self.duration))

        # Number of test cycles
        row += 1
        lbNo = tkinter.Label(
            self.main, text=_("Number of cycles"))
        lbNo.grid(row=row, column=0, sticky="w")

        self.enNo = tkinter.Entry(
            self.main, width=entrywidth)
        self.enNo.grid(row=row, column=1, columnspan=3)
        self.enNo.insert(0, str(self.no))

        # List of scale factors
        row += 1
        lbScales = tkinter.Label(
            self.main, text=_("List of scales"))
        lbScales.grid(row=row, column=0, sticky="w")

        self.enScales = tkinter.Entry(
            self.main, width=entrywidth)
        self.enScales.grid(row=row, column=1, columnspan=3)
        out = ""
        for scale in self.scales:
            out += " " + str(scale)
        self.enScales.insert(0, out)

        # Prepare transactions
        row += 1
        lbPrepare = tkinter.Label(
            self.main, text=_("Prepare transactions"))
        lbPrepare.grid(row=row, column=0, sticky="w")

        self.ivPrepare = tkinter.IntVar()
        chPrepare = tkinter.Checkbutton(
            self.main, variable=self.ivPrepare,
            onvalue=True, offvalue=False)
        chPrepare.grid(row=row, column=1)
        if self.prepare:
            chPrepare.select()
        else:
            chPrepare.deselect()

        # Remove database after tests
        row += 1
        lbRemove = tkinter.Label(
            self.main, text=_("Remove database"))
        lbRemove.grid(row=row, column=0, sticky="w")

        self.ivRemove = tkinter.IntVar()
        chRemove = tkinter.Checkbutton(
            self.main, variable=self.ivRemove,
            onvalue=True, offvalue=False)
        chRemove.grid(row=row, column=1)
        if self.remove:
            chRemove.select()
        else:
            chRemove.deselect()

        # Reconnect for each transaction
        row += 1
        lbReconnect = tkinter.Label(
            self.main, text=_("Reconnect per transaction"))
        lbReconnect.grid(row=row, column=0, sticky="w")

        self.ivReconnect = tkinter.IntVar()
        chReconnect = tkinter.Checkbutton(
            self.main, variable=self.ivReconnect,
            onvalue=True, offvalue=False)
        chReconnect.grid(row=row, column=1)
        if self.reconnect:
            chReconnect.select()
        else:
            chReconnect.deselect()

        # Retry failed transactions
        row += 1
        lbRetry = tkinter.Label(
            self.main, text=_("Retry failed transactions"))
        lbRetry.grid(row=row, column=0, sticky="w")

        self.ivRetry = tkinter.IntVar()
        chRetry = tkinter.Checkbutton(
            self.main, variable=self.ivRetry,
            onvalue=True, offvalue=False)
        chRetry.grid(row=row, column=1)
        if self.retry:
            chRetry.select()
        else:
            chRetry.deselect()

        # Buttons
        row += 1
        btOK = tkinter.Button(
            self.main, text=_("OK"), command=self.ok)
        btOK.grid(row=row, column=0)

        btIgnore = tkinter.Button(
            self.main, text=_("Ignore"), command=self.ignore)
        btIgnore.grid(row=row, column=2)

        btCancel = tkinter.Button(
            self.main, text=_("Cancel"), command=self.cancel)
        btCancel.grid(row=row, column=3)

        # Start event loop
        self.main.mainloop()
        return self.retval

    def results(self):
        """Pickup results.

        Returns:
            A tuple containing the test conditions. See Args of constructor
            for details!
        """
        return (self.db, self.transtype, self.output, self.nodes,
                self.threads, self.clients, self.duration, self.no,
                self.scales, self.prepare, self.remove, self.reconnect,
                self.retry)


if __name__ == "__main__":
    """Testing interface.

    The main routine is used to test the graphical interface.
    """
    z = tpcbWin("PG", TPCB_DEFAULT, "results", 1, 2, 4, 60.0, 10,
                (1, 2, 5, 10),
                False, True, False, False)
    if z.show():
        print(z.results())
    else:
        print("Cancel pressed")
