# -*- coding: utf-8 -*-
#
# dsbench/tpcbNode.py - Remote processing
#
# Copyright (C) 2014-2015 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

import socketserver

from tpcbCore import *
from tpcbPg import *
from tpcbFdb import *
from tpcbFirebirdSql import *
# from tpcbOra import *


# i18n preparations.
def _(s):
    return s


class tpcbNode(socketserver.UDPServer):
    """UDP server.

    This class implements a non threaded UDP server that only catches
    broadcast messages from master to locate nodes.
    """
    def __init__(self, env, addr, handler):
        """Initialisation.

        Args:
            env: Test environment, a tpcbEnv() instance.
            addr: Address and port the UDP server should listen to.
            handler: A tpcbNodeHandler() class that act as handler if the
                master sends a message.
        """
        socketserver.UDPServer.__init__(self, addr, handler)
        self.env = env
        self.stop = False

    def serve_forever(self):
        """Start server.

        This is an override of the main listening loop. This loop is able
        to abort if stop flag is set by the handler after receiving a
        killing instruction.
        """
        while True:
            self.handle_request()
            if self.stop:
                break


class tpcbNodeHandler(socketserver.BaseRequestHandler):
    """UDP request handler.

    This request handler class is responsible for intercepting UDP
    broadcasts and starting transaction jobs on a remote node.
    """
    def close(self):
        """Close connection to master."""
        self.sock.close()
        self.logger.debug(_("Disonnected from master '%s'."), self.addr)

    def error(self, msg, *args):
        """Handle error.

        Log an error message and then close connection to the master.

        Args:
            msg: Message string to be logged.
            *args: Positional arguments to be inserted in the message.
        """
        self.logger.error(msg, *args)
        self.close()

    def handle(self):
        """Handler.

        This method is invoked if a UDP broadcast came in.
        It checks the message, connects to the master, sends an
        identification and waits until a job is delivered by the master.
        If a job is injected it will be carried out and the results are
        send back to the master. This handler is called non threaded this
        means it can run only once.
        """
        # Initialisation
        env = self.server.env
        self.logger = env.logger

        # Extract broadcast message from master
        msg = tpcbMsg(buffer=self.request[0])
        if msg.isscan:
            pass
        elif msg.iskill:
            self.logger.info(_("Kill request received. Terminate now."))
            self.server.stop = True
            return
        else:
            self.logger.error(_("Wrong broadcast request '%s'."), str(msg))
            return

        self.logger.info(
            _("Incoming broadcast from master '%s'."),
            str(self.client_address[0]))

        # -------------------------------------------------------
        # For testing purposes:
        # Do not connect to master in some cases to test reaction
        # of the master to node errors or interrupts.
        # if random.randint(0, 20) <= 1:
        #    self.logger.info(_("Close node for testing purposes."))
        #    return
        # -------------------------------------------------------

        # Open a TCP socket to the master
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        try:
            # Connect to master
            self.sock.settimeout(TIMEOUT)
            self.addr = (self.client_address[0], env.port)
            self.sock.connect(self.addr)
            self.logger.debug(_("Connected to master '%s'."), self.addr[0])

            # Send a reply containing node's host specification
            env.own.ip = self.client_address[0]
            msg = tpcbMsg(msgtype="NODE", parameters=str(env.own))
            self.logger.debug(
                _("Send broadcast response to master '%s': '%s'."),
                self.client_address[0], str(msg))
            self.sock.sendall(bytes(msg))

            # Wait for a job from master for some time
            buffer = self.sock.recv(BUFSIZE)
            if not buffer:
                self.error(_("Connection closed by master '%s'. Abort now."),
                           self.client_address[0])
                return
            msg = tpcbMsg(buffer=buffer)
            self.logger.debug(_("Message received from master '%s': '%s'."),
                              self.client_address[0], str(msg))
            if not msg.isjob:
                self.error(
                    _("Message from master '%s' is not a valid job request."),
                    self.client_address[0])
                return

            env.job = eval(msg.parameters)

            # Special case: "localhost"
            # This means localhost on master, hence replace
            # server localhost by the IP address of the master, which
            # is: self.client_address[0].
            env.replace_localhost(self.client_address[0])

            # Output a test environment representation
            self.logger.debug(_("Environment: '%s'."), repr(env))

            # Analyse and execute the job
            m = tpcbSingle(env, env.job.scale)
            if env.job.db == "PG":
                db = tpcbDbPostgres(env, None)
            elif env.job.db == "FB" and env.job.fbdriver == "fdb":
                db = tpcbDbFdb(env, None)
            elif env.job.db == "FB":
                db = tpcbDbFirebirdSql(env, None)
            # elif env.job.db == "ORA":
            #    db = tpcbDbOracle(env, None)

            env.blankinfo(
                _("--- Execute transaction cycle #{0:d} for scale {1:d} "
                  "on '{2:s}' database ---").format(env.job.cycle,
                                                    env.job.scale, db.name))

            try:
                # Run the job
                db.create(m, False)
                results = db.run_local()
                db.drop(False)
                # Send result
                result = tpcbMsg(msgtype="RESULT", parameters=str(results))
                self.sock.sendall(bytes(result))
                # Send end mark
                time.sleep(SENDDELAY)
                endmark = tpcbMsg(msgtype="ENDMARK")
                self.sock.sendall(bytes(endmark))
                # Send single transaction times
                time.sleep(SENDDELAY)
                times = tpcbMsg(buffer=m.buf.tobytes())
                self.sock.sendall(bytes(times))
                # Send end mark
                time.sleep(SENDDELAY)
                endmark = tpcbMsg(msgtype="ENDMARK")
                self.sock.sendall(bytes(endmark))
            except Exception as e:
                self.error(
                    _("Could not execute job! Error Message: '%s'."),
                    str(e))
            finally:
                del db
                del m

        except socket.timeout:
            self.error(_("Did not start transaction job "
                         "due to communication timeout."))

        except ConnectionResetError:
            self.logger.warn(_("Connection closed by peer."))


class tpcbNodeLauncher():
    """Node launcher.

    This class contains the main server loop for the UDP server.
    """
    def __init__(self, env):
        """Initialisation.

        Args:
            env: Test environment, a tpcbEnv() instance.
        """
        self.env = env
        self.env.logger.info(_("Launch application as remote node."))

    def start(self):
        """Start server.

        This methos installs a non threaded UDP server and starts the main
        listening loop.
        """
        addr = ("", self.env.port)
        self.env.logger.info(
            _("Install UDP server listening on '%s'."), addr)
        server = tpcbNode(self.env, addr, tpcbNodeHandler)
        server.serve_forever()
        del server


class tpcbNodeKiller():
    """Node killer.

    This is a helper for killing all nodes in the LAN.
    """
    def __init__(self, env):
        """Initialisation.

        Args:
            env: Test environment, a tpcbEnv() instance.
        """
        self.env = env
        self.env.logger.info(_("Kill all nodes requested."))

    def broadcast(self):
        """Send kill request.

        This method sends a UDP broadcast to kill all nodes in the LAN.
        """
        self.env.logger.info(_("Send UDP broadcast to kill nodes."))
        s = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, True)
        msg = tpcbMsg(msgtype="KILL")
        s.sendto(bytes(msg), ("<broadcast>", self.env.port))
        s.close()
