# -*- coding: utf-8 -*-
#
# dsbench/tpcbPg.py - PostgreSQL driver
#
# Copyright (C) 2014-2015 Rajko Thon <rajko.thon@leitstern.de>
#                         Mathias Ball <meb@leitstern.de>
#
# DSBench is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# You must obey the GNU Lesser General Public License in all respects for
# all of the code used.
#
# DSBench is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#

from tpcbCore import *
from tpcbDbBase import *


# i18n preparations.
def _(s):
    return s


class tpcbDbPostgres(tpcbDb):
    """Database controller for PostgreSQL.

    This controller overrides tpcbDb() class method for accessing a
    PostgreSQL database.
    """
    def __init__(self, env, multiple):
        """Initialisation.

        Args:
            env: The test environment, a tpcbEnv() instance.
            multiple: The measurements table, a tpcbMultiple() instance.
        """
        if not PSYCOPG2:
            raise Exception(_("Python module 'psycopg2' not installed."))

        tpcbDb.__init__(self, env, multiple)
        self.dbname = "tpc"
        self.server = env.job.pgServer
        self.user_name = "tpc"
        self.schema_name = "main"
        self.password = "tpc"
        self.name = "PostgreSQL"
        self.basepath = "postgres"
        self.conn = "dbname='{0:s}' user='{1:s}' host='{2:s}' " \
                    "password='{3:s}'".format(self.dbname, self.user_name,
                                              env.job.pgServer, self.password)
        self.postgres = "dbname='postgres' user='postgres' " \
                        "host='{0:s}' password='{1:s}'".format(
                            env.job.pgServer, env.job.pgPassword)

    def _connect(self):
        """Connect to the database.

        This is an override method for the PostgreSQL database.

        Returns:
            A connection and a cursor to the database.
        """
        con = None
        if not self.env.job.reconnect:
            self.env.logger.debug(
                _("Connect to PostgreSQL database '%s' on '%s'."),
                self.dbname, self.server)
        try:
            con = psycopg2.connect(self.conn)
            con.autocommit = True
            cur = con.cursor()
            cur.execute("set search_path to " + self.schema_name)
            if not self.env.job.reconnect:
                self.env.logger.debug(
                    _("Set search_path to '%s'."), self.schema_name)
            return con, cur
        except psycopg2.Error as e:
            self.env.logger.exception(
                _("Exception while connecting PostgreSQL database: "
                  "pgcode='%s', pgerror='%s'."), str(e.pgcode), str(e.pgerror))
            raise e

    def _disconnect(self, con, cur):
        """Disconnect from database.

        This is an override method for the PostgreSQL database.

        Args:
            con: Connection to the database
            cur: Cursor from the database connection
        """
        if not self.env.job.reconnect:
            self.env.logger.debug(
                _("Disconnect from PostgreSQL database '%s' on '%s'."),
                self.dbname, self.server)
        cur.close()
        con.close()

    def _create(self, scale):
        """Create a database.

        This is an override method for the PostgreSQL database.
        An extisting database having the same name will be deleted!

        Args:
            scale: Scaling factor of the database.

        Returns:
            A tuple containing durations in seconds for filling the
            database, creating the primary and foreign keys.
        """
        self.env.logger.info(
            _("Recreating PostgreSQL database '%s' on '%s' with scale=%d."),
            self.dbname, self.server, scale)

        # Drop old database
        con = None
        con = psycopg2.connect(self.postgres)
        con.autocommit = True
        cur = con.cursor()

        cur.execute("drop database if exists " + self.dbname)
        self.env.logger.debug(_("Database '%s' dropped."), self.dbname)
        cur.execute("drop user if exists " + self.user_name)
        self.env.logger.debug(_("User '%s' deleted."), self.dbname)
        cur.execute("create user " + self.user_name +
                    " password '" + self.user_name + "'")
        self.env.logger.debug(_("User '%s' created."), self.user_name)
        self.fileexec(_("Database {0:s} created.").format(self.dbname),
                      "create_database.sql", con, cur)

        cur.close()
        con.close()

        # Schema
        con = None
        con = psycopg2.connect(self.conn)
        con.autocommit = True
        cur = con.cursor()

        cur.execute("drop schema if exists " + self.schema_name)
        self.env.logger.debug(_("Schema '%s' dropped."), self.schema_name)
        cur.execute("create schema " + self.schema_name +
                    " authorization " + self.user_name)
        cur.execute("comment on schema " + self.schema_name +
                    " is 'TPC-B test database'")
        self.env.logger.debug(_("Schema '%s' created."), self.schema_name)
        cur.execute("grant all on database " + self.dbname +
                    " to " + self.user_name)
        self.env.logger.debug(_("Database grants to '%s'."), self.user_name)
        cur.execute("grant all on schema " + self.schema_name +
                    " to " + self.user_name)
        self.env.logger.debug(_("Schema grants to '%s'."), self.user_name)
        cur.execute("set search_path to " + self.schema_name)
        self.env.logger.debug(_("Set search-path to '%s'."), self.schema_name)

        cur.close()
        con.close()

        # Tables
        con, cur = self._connect()
        self.fileexec(_("Table 'tpcb_branch' created."),
                      "tpcb_branch.sql", con, cur)
        self.fileexec(_("Table 'tpcb_teller' created."),
                      "tpcb_teller.sql", con, cur)
        self.fileexec(_("Table 'tpcb_account' created."),
                      "tpcb_account.sql", con, cur)
        self.fileexec(_("Table 'tpcb_history' created."),
                      "tpcb_history.sql", con, cur)

        # Functions
        self.fileexec(_("Function 'func_init' created."),
                      "func_init.sql", con, cur)
        self.fileexec(_("Function 'func_work' created."),
                      "func_work.sql", con, cur)
        self.fileexec(_("Function 'func_work_select' created."),
                      "func_work_select.sql", con, cur)

        # Initialisation
        tInit = self.timeexec(
            _("Initializing TPC-B with scale={0:d}.").format(scale),
            "select tpcb_func_p_init(1, 10, 100000, {0:d})".format(scale),
            con, cur)

        # Create index
        tPk = 0.0
        tFk = 0.0
        # Branch
        tPk += self.timeexec(
            _("Creating primary key on 'tpcb_branch'."),
            "alter table tpcb_branch add constraint tpcb_branch_pk "
            "primary key (branch_id)",
            con, cur)

        # Teller
        tPk += self.timeexec(
            _("Creating primary key on 'tpcb_teller'."),
            "alter table tpcb_teller add constraint tpcb_teller_pk "
            "primary key (teller_id)",
            con, cur)
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_branch' on 'tpcb_teller'."),
            "alter table tpcb_teller add constraint tpcb_teller_fk "
            "foreign key (branch_id) references tpcb_branch(branch_id)",
            con, cur)

        # Account
        tPk += self.timeexec(
            _("Creating primary key on 'tpcb_account'."),
            "alter table tpcb_account add constraint tpcb_account_pk "
            "primary key (account_id)",
            con, cur)
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_branch' on 'tpcb_account'."),
            "alter table tpcb_account add constraint tpcb_account_fk "
            "foreign key (branch_id) references tpcb_branch(branch_id)",
            con, cur)

        # History
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_branch' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_1 "
            "foreign key (branch_id) references tpcb_branch(branch_id)",
            con, cur)
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_teller' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_2 "
            "foreign key (teller_id) references tpcb_teller(teller_id)",
            con, cur)
        tFk += self.timeexec(
            _("Creating foreign key to 'tpcb_account' on 'tpcb_history'."),
            "alter table tpcb_history add constraint tpcb_history_fk_3 "
            "foreign key (account_id) references tpcb_account(account_id)",
            con, cur)

        self.env.logger.debug("t(INIT)=%.2f, t(PK)=%.2f, t(FK)=%.2f",
                              tInit, tPk, tFk)

        self._disconnect(con, cur)
        return tInit, tPk, tFk

    def _drop(self):
        """Drop the database.

        This is an override method for the PostgreSQL database.

        Returns:
            Time needed to drop the database in seconds.
        """
        self.env.logger.info(_("Drop PostgreSQL database '%s'."), self.dbname)

        con = None
        con = psycopg2.connect(self.postgres)
        con.autocommit = True
        cur = con.cursor()

        tDrop = self.timeexec(
            _("Drop database '{0:s}'.").format(self.dbname),
            "drop database if exists {0:s}".format(self.dbname),
            con, cur)
        tDrop += self.timeexec(
            _("Drop user '{0:s}'.").format(self.user_name),
            "drop user if exists {0:s}".format(self.user_name),
            con, cur)

        self.env.logger.debug("t(DROP)=%.2f", tDrop)

        cur.close()
        con.close()
        return tDrop

    def _clean(self):
        """Clean the database between measurement cycles.

        This is an override method for the PostgreSQL database.
        It uses the same vacuum run which is standard in "pgbench".

        Returns:
            Time needed to clean the database in seconds.
        """
        con, cur = self._connect()

        tClean = 0.0
        tClean += self.timeexec(_("Vacuum 'tpcb_branch'."),
                                "vacuum tpcb_branch", con, cur)
        tClean += self.timeexec(_("Vacuum 'tpcb_teller'."),
                                "vacuum tpcb_teller", con, cur)
        tClean += self.timeexec(_("Truncate 'tpcb_history'."),
                                "truncate tpcb_history", con, cur)

        self._disconnect(con, cur)
        self.env.logger.debug("t(CLEAN)=%.2f", tClean)
        return tClean

    def _size(self):
        """Determine the physical size of the database in GB.

        This is an override method for the PostgreSQL database.

        Returns:
            The size of the database in GB as a float.
        """
        con = None
        con = psycopg2.connect(self.postgres)
        cur = con.cursor()
        sql = "select pg_database_size('{0:s}');".format(self.dbname)
        cur.execute(sql)
        row = cur.fetchone()
        cur.close()
        con.close()
        return row[0] / GB

    def _execute(self, sql, con, cur):
        """Executes SQL commands.

        This is an override method for the PostgreSQL database.

        Args:
            sql: SQL statements.
            con: Connection to the database.
            cur: Cursor on the connection.
        """
        cur.execute(sql)

    def prepare(self, cur):
        """Prepare.

        Makes a preparation for the TPC-B transaction statement.

        Args:
            cur: Cursor on the connection.
        """
        sql = "prepare transaction (bigint, bigint, bigint, bigint) " \
              "as select {0:s}($1, $2, $3, $4)".format(
                  tpcbTransTable[self.env.job.transtype].func)
        cur.execute(sql)

    def command(self, branch_id, teller_id, account_id, delta):
        """Transaction statement.

        Creates the SQL command for one TPC-B transaction.
        Delivers different output depending on prepare is used.

        Args:
            branch_id: ID of branch.
            teller_id: ID of teller.
            account_id: ID of account.
            delta: Amount.

        Returns:
            An SQL statement as string.
        """
        if self.env.job.prepare:
            sql = "execute transaction({0:d}, {1:d}, {2:d}, {3:d})".format(
                branch_id, teller_id, account_id, delta)
        else:
            sql = "select {0:s}({1:d}, {2:d}, {3:d}, {4:d})".format(
                tpcbTransTable[self.env.job.transtype].func,
                branch_id, teller_id, account_id, delta)
        return sql

    def transaction(self, sql, cur):
        """Execute transaction.

        Executes a single transaction including return value pickup.

        Args:
            sql: SQL statements.
            cur: Cursor on the connection.
        """
        cur.execute(sql)
        x = cur.fetchone()[0]

    def chkerror(self, e):
        """Error check.

        Checks if the exception is caused by a temporary database error
        that is normal operation.

        Args:
            e: Exception to be checked.

        Returns:
            A boolean that is False if the error is normal operation.
        """
        if not isinstance(e, psycopg2.extensions.TransactionRollbackError):
            return True
        else:
            return False

    def _loop_single(self, evkill):
        """Transaction loop with single connection.

        This is an override method for the PostgreSQL database.
        Each transaction opens and closes the database connection.

        Args:
            evkill: A thread event to signalise kill.

        Returns:
            A tuple containing the number of successful transactions done,
            the number of rejected transactions due to collisions, the total
            transaction residence time and effective duration elapsed during
            transactions.
        """
        finish = False
        residence = 0.0
        elapsed = 0.0
        done = 0
        errors = 0
        t0 = timeit.default_timer()

        # Run transaction's cycle
        while True:
            # Random numbers for transaction input values
            branch_id, teller_id, account_id, delta = self.rand.next()

            # Open connection and prepare transaction
            t1 = timeit.default_timer()
            con, cur = self._connect()
            self.prepare(cur)
            sql = self.command(branch_id, teller_id, account_id, delta)

            # Do transaction
            while True:
                ok = True
                try:
                    self.transaction(sql, cur)
                    self._disconnect(con, cur)
                except Exception as e:
                    # Catch errors while transactions
                    if self.chkerror(e):
                        raise e
                    ok = False
                finally:
                    # Get current time and check duration
                    t2 = timeit.default_timer()
                    if t2 - t0 > self.env.job.duration:
                        # Mark as duration finished, waste transaction
                        # and break the inner retry loop
                        finish = True
                        break
                    else:
                        # Register effective duration
                        elapsed = t2 - t0
                        if ok:
                            # Transaction succeeded
                            # Register transaction residence time
                            diff = t2 - t1
                            self.single.buf.set(diff)
                            # Sum total residence time
                            residence += diff
                            done += 1
                            break
                        else:
                            # Transaction failed
                            errors += 1
                            # Retrying transaction until success
                            if not self.env.job.retry:
                                break

            # Abort loop due to end of duration
            if finish:
                break

            # Abort loop if thread receives kill request
            if evkill and evkill.isSet():
                break

        return done, errors, residence, elapsed

    def _loop_multi(self, evkill, cons, curs):
        """Transaction loop with multiple connections.

        This is an override method for the PostgreSQL database.
        Connections are opened and closed only once for the whole
        measurement cycle.

        Args:
            evkill: A thread event to signalise kill.
            cons: A list of connections to the database.
            curs: A list of cursors on the connections.

        Returns:
            A tuple containing the number of successful transactions done,
            the number of rejected transactions due to collisions, the total
            transaction residence time and effective duration elapsed during
            transactions.
        """
        # Prepare
        if self.env.job.prepare:
            for cur in curs:
                self.prepare(cur)

        # Initialise cycle
        c = 0
        cmax = len(curs)
        finish = False
        residence = 0.0
        elapsed = 0.0
        done = 0
        errors = 0
        t0 = timeit.default_timer()

        # Run transaction's cycle
        while True:
            # Use clients via "round robin"
            c = (c + 1) % cmax
            # Random numbers for transaction input values
            branch_id, teller_id, account_id, delta = self.rand.next()
            sql = self.command(branch_id, teller_id, account_id, delta)

            # Do transaction
            t1 = timeit.default_timer()
            while True:
                ok = True
                try:
                    self.transaction(sql, curs[c])
                except Exception as e:
                    # Catch errors while transactions
                    if self.chkerror(e):
                        raise e
                    ok = False
                finally:
                    # Get current time and check duration
                    t2 = timeit.default_timer()
                    if t2 - t0 > self.env.job.duration:
                        # Mark as duration finished, waste transaction
                        # and break the inner retry loop
                        finish = True
                        break
                    else:
                        # Register effective duration
                        elapsed = t2 - t0
                        if ok:
                            # Transaction succeeded
                            # Register transaction residence time
                            diff = t2 - t1
                            self.single.buf.set(diff)
                            # Sum total residence time
                            residence += diff
                            done += 1
                            break
                        else:
                            # Transaction failed
                            errors += 1
                            # Retrying transaction until success
                            if not self.env.job.retry:
                                break

            # Abort loop due to end of duration
            if finish:
                break

            # Abort loop if thread receives kill request
            if evkill and evkill.isSet():
                break

        return done, errors, residence, elapsed

    def _parameters(self):
        """Collect database parameters.

        This is an override method for the PostgreSQL database.
        Parameters are queried from the database itself.
        """
        con = None
        con = psycopg2.connect(self.postgres)
        con.autocommit = True
        cur = con.cursor()

        # "effective_cache_size"
        cur.execute("select setting from pg_settings where "
                    "name='effective_cache_size';")
        row = cur.fetchone()
        effective = float(row[0]) / 128.0 / 1024.0

        # "shared_buffers"
        cur.execute("select setting from pg_settings where "
                    "name='shared_buffers';")
        row = cur.fetchone()
        shared = float(row[0]) / 128.0 / 1024.0

        # "work_mem"
        cur.execute("select setting from pg_settings where name='work_mem';")
        row = cur.fetchone()
        work = float(row[0]) / 1024.0 / 1024.0

        # "transaction_isolation"
        cur.execute("select setting from pg_settings where "
                    "name='transaction_isolation';")
        isolation = cur.fetchone()[0]

        # data_directory
        cur.execute("select setting from pg_settings where "
                    "name='data_directory';")
        data_directory = cur.fetchone()[0]

        # General parameters
        server_version = con.server_version

        cur.close()
        con.close()

        self.env.dbpath = data_directory

        self.env.dbpar("PostgreSQL version", server_version)
        self.env.dbpar("effective_cache_size [GB]", effective)
        self.env.dbpar("shared_buffers [GB]", shared)
        self.env.dbpar("work_mem [GB]", work)
        self.env.dbpar("transaction_isolation", isolation)
        self.env.dbpar("data_directory", data_directory)
        self.multiple.setmarker(work, "work_mem")
        self.multiple.setmarker(shared, "shared_buffer")
        self.multiple.setmarker(effective, "effective_cache_size")
